%hybrid_sys simulates a hybrid system
%[ TH,X,OP,TJ,XJ ]=hybrid_sys( Flowmap, Jmpmap, Jmpcon, OPmap, TermId, Pat, t0, tf, x0, opt, Params, Add_par)
%
%Flowmap: Function handle to an ode function for the flow of the system
%Jmpmap:  Function handle to a function for jumps in the sytem. The format
%         for defining the jump map is as follows.
%
%         xnew = jump_map( th, x, Params, jump_flag )
%         th - current hybrid time
%         x  - current state (column vector)
%         Params - see below
%         jump_flag - a vector of jump flags returned by Jmpcon
%         xnew - new state obtained after applying the jump map
%
%Jmpcon:  Function handle to a function that determines when to jump. The
%         system is assumed to flow by default and a jump occurs whenever
%         the JumpCon is satisfied. The function must be an event function
%         that can be used with ode45. The length of the value vector
%         returned by this function must be equal to P, where P is the
%         number of elements in the cell array Pat.
%OPmap:   Function handle to a function to compute the outputs of the
%         system. The format for defining the OPmap is as follows.
%
%         OP = op_map( TH, X, TJ, XJ, Params )
%         OP can be a cell array or a matrix
%
%TermId:  Vector of ids of the jump flags, which when set the integration
%         terminates. The values can be integers between 1 and the number
%         of jump flags.
%Pat:     Pat is a cell array of P vectors. The value vector returned by
%         the Jumpcon function must be of length P. The vector Pat{i}
%         must contain non-repeating numbers from 1 to N, where N is the
%         number of states of the system. See TJ and XJ for more.
%t0:      Initial continuous-time at which the simulation starts
%tf:      Continuous-time at which the simulation ends
%x0:      Initial state of the system
%opt:     Options to be used with the integration function ode45. Must be
%         set using odeset function.
%Params:  Structure of additional Params required by Flowmap, Jumpmap,
%         Jumpcon, OPmap
%Add_Par: OPtional. Additional parameters that user can set. Provide a
%         single structure with all the additional parameters to be set.
%         Currently supported fieldnames of the structure:
%            MJTol: If there are multiple jump conditions, when one of the
%            jump conditions is satisfied, another jump condition may be
%            close to being satisfied (which as a result may be missed
%            subsequently). Thus, whenever any of the conditions "jumps",
%            all the other conditions whose values at the time are within
%            MJTol of zero in the relevant direction. Must be positive.
%            Default value is 1e-12.
%
%Returns:
%
%TH:      Hybrid time of the solution data. The first column contains the
%         continuous-time and the second column contains the jump-time.
%         TH is such that a row differs from its previous row in exactly
%         one of the columns
%X:       State solution as a function of hybrid time TH. Has same number
%         of rows as TH.
%TJ:      Cell array of P vectors, where P is the number of elements in the
%         cell array Pat. TJ{i} is a vector containing the sequence of
%         continuou-times at which i^th element in the value vector
%         returned by Jmpcon denotes a Jump.
%XJ:      Cell array of P vectors. XJ{i} contains the sequence of values
%         the states listed in P{i} take just after jumps at
%         continuous-times given in TJ{i}
%OP:      Structure containing additional outputs of the system,
%         determined by calling the OPmap function. Note that OP is
%         computed last and can depend on TH, X, TJ, XJ, Params
%
%Updated on 2014/7/7: multiple consecutive jumps now possible
%
%Author: Pavan Tallapragada

function [ TH,X,TJ,XJ,OP ]=hybrid_sys( Flowmap, Jmpmap, Jmpcon, OPmap, ...
                                       TermId, Pat, t0, tf, x0, opt, ...
                                       Params, varargin )

%Tolerance for determining closely spaced events generated by separate jump
%conditions. See line " jf = jump_flag; " and the one following it for more
%details
MJTol = 1e-12;
if nargin > 11
    ad_params_set = 0;
    if isfield(varargin{1}, 'MJTol')
        MJTol = abs(varargin{1}.MJTol);
        disp('Tolerance for simulataneous jumps of separate jump conditions:');
        disp(MJTol);
        ad_params_set = 1;
    end
    if ad_params_set == 0
        disp('No params set');
    end
end

N = length(x0);
P = length(Pat);

Samp = 20000;
TH = zeros(Samp,2); X = zeros(Samp,N); OP = [ ];

for i = 1:P
    TJ{i} = zeros(Samp,1);
    XJ{i} = zeros(Samp,length(Pat{i}));
    Pat{i} = sort(Pat{i});
    if sum( ( Pat{i}(2:end) - Pat{i}(1:end-1) ) == 0 ) > 0
        disp('Check Pat, some non-unique elements in one of the vectors');
        return;
    end
end
NJmps = zeros(P,1);

TH(1,:) = [t0, 0]; X(1,:) = x0';

ip = 0; in = 1; % i_previous and i_new
terminate = 0;

while TH(in,1)~=tf
    
    %check jumps at t = TH(in,1) and keep jumping until the state does not
    %go back to the flow set.
    
    while(1)        
        [jump_flag,isterm,direction] = Jmpcon(TH(in,1), X(in,:)', Params);

        %if some jump flags are close to be set but not yet set. Then set
        %them now so that they are not missed in the next run of ode45
        jf = jump_flag;
        if direction == 1
            jump_flag =  jf.*(jf < -MJTol | jf >= 0) + ...
                         zeros(size(jf)).*(jf >= -MJTol & jf <= 0);
        elseif direction == -1
            jump_flag =  jf.*(jf > MJTol | jf <= 0) + ...
                         zeros(size(jf)).*(jf >= 0 & jf <= MJTol);
        elseif direction == 0
            jump_flag =  jf.*(jf < -MJTol | jf > MJTol) + ...
                         zeros(size(jf)).*(jf >= -MJTol & jf <= MJTol);
        end
        if length(jump_flag) ~= P
            disp('Jmpcon not returning a vector of correct length');
            return;
        end

        if sum(jump_flag(TermId) >= 0) > 0
            terminate = 1;
            break;
        end

        if sum(jump_flag >= 0) > 0
            %if at least one of the jump flags is set then
            TH(in+1,1) = TH(in,1);
            TH(in+1,2) = TH(in,2) + 1;
            X(in+1,:) = ( Jmpmap(TH(in,:),X(in,:)', Params, jump_flag ) )';

            in = in+1;
        else
            break;
        end

        for i = 1:P
            if jump_flag(i) >= 0    %if need to sample
                NJmps(i) = NJmps(i) + 1;
                TJ{i}(NJmps(i)) = TH(in,1);
                XJ{i}(NJmps(i),:) = X(in,Pat{i});
            end
        end
        
    end
    
    if terminate == 1
        break;
    end
    
    %%%%%%%%%%%% Integration till next jump %%%%%%%%%%%%%%%
        
    [tn, xn] = ode45(Flowmap,[ TH(in,1), tf ], X(in,:)', opt, Params);
    
    %%%%%%%%%% Data storage %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
    
    ip = in;
    in = length(tn) + ip - 1;
    
    TH(ip:in,1) = tn; TH(ip:in,2) = TH(ip,2);
    X(ip:in,:) = xn;
end

%Retain only relevant part of the matrices
TH = TH(1:in,:); X = X(1:in,:);
for i = 1:P
    TJ{i} = TJ{i}(1:NJmps(i));
    XJ{i} = XJ{i}(1:NJmps(i),:);
end

OP = OPmap( TH, X, TJ, XJ, Params );